# SPEC_FILE = math.sp;
# LIST_FILE = math.lst;
# MASK_COUNT = 8;
# SET_ADDRESS = 0;
# LOAD_ADDRESS = 0;

/*
 * Memory map: Each page is 256 bytes.
 *
 * Page 0:
 * 		00-31	matrix A
 * 		32-63	matrix B
 * 		64-95	matrix C
 *
 * 		new transformation matrices are written into B, and
 * 		the CTM swaps between A and C (to avoid one copy)
 *
 * 		096	&CTM
 *		097	&NCTM
 * 		098-105	TMP1-TMP8
 *
 * 		106-111 eye point
 * 		112-117 fixatoint point
 *
 * 		118	number of vertices in object (in page 1)
 * 		119	number of edges in object
 * 		120-255	edge list
 *
 * 		68 edges (pairs of integers) can be stored
 *
 * Page 1:
 * 		00-255	matrix of vertices
 *
 * 		32 vertices can be stored
 *
 * Page 2:
 * 		00-255	transformed matrix of vertices
 */

start:
        /*
	 * Initialization:
	 * 1) set up default CTM, object, etc (ROM dump)
	 * 2) set TMP1-TMP8 to zero
	 * 3) reset ACK_O and DAV_O
	 */
	jmp main_loop;

main_loop:
	/* read a block opcode into B */
	LDA 1;
	LDMAR TMP1;
	LDREG r4 l5;
	jmp read_bytes;
l5:	LDB DRREG r1;

	/* Jump to the appropriate block handler.  The order here */
	/* depends on the opcode values defined by the protocol. */
	LDA OBJECT;
	latch;
	if /equal cjmp not_object;
	jmp object_block;

not_object:
	LDA aplus1;
	if /equal cjmp not_eye;
	jmp eye_block;

not_eye:
	LDA aplus1;
	if /equal cjmp not_fix;
	jmp fix_block;

not_fix:
	LDA aplus1;
	if /equal cjmp not_identify;
	jmp identify_block;

not_identify:
	LDA aplus1;
	if /equal cjmp not_transform;
	jmp transform_block;

not_transform:
	/* the block opcode is not defined */
	jmp bad_block;

/***********************************************************************
 * check_eob_marker -- read a byte, check that it is 0xff, and return
 * to the main loop.  This is used as a return addr from read_bytes.
 ***********************************************************************/
check_eob_marker:
	LDB %hff;
	LDMAR TMP1;
	LDA 1;
	LDREG r4 check_eob_marker_1;
	jmp read_bytes;
check_eob_marker_1:
	latch;
	if /equal cjmp bad_block;
	jmp main_loop;

/***********************************************************************
 * identify_block -- read nothing, reset CTM to identity (well...)
 ***********************************************************************/
identify_block:
	jmp check_eob_marker;

/***********************************************************************
 * fix_block, eye_block -- read three bytes into EYE_POINT or FIX_POINT
 ***********************************************************************/
eye_block:
	LDMAR EYE_POINT;
	jmp eye_or_fix_block;

fix_block:
	LDMAR FIX_POINT;
	jmp eye_or_fix_block;

eye_or_fix_block:
	LDREG r4 check_eob_marker;
	LDA 3;
	jmp read_bytes;

/***********************************************************************
 * transform_block -- read in a transformation matrix, apply it to the 
 * CTM, and render the new image (well...)
 ***********************************************************************/
transform_block:
	/* read all transformations in matrix B */
	LDMAR MATRIX_B;
	LDREG r4 transform_block_rtn_1;
	/* load 5 to loop 4 times, since we will compare against 1 */
	LDREG r2 5;
transform_block_1:
	LDA 4;
	jmp read_bytes;
transform_block_rtn_1:
	/* read 4 bytes, make sure the last one is 1 */
	LDA DRREG r1;
	latch;
	if /equal cjmp bad_block;

	/* row was valid, see if we've read four rows */
	LDA DRREG r2;
	latch;
	if /equal cjmp transform_block_1;

	/* check end-of-block marker; A is already 1 */
	LDMAR TMP1;
	LDREG r4 transform_block_rtn_2;
	jmp read_bytes;
transform_block_rtn_2:
	/* 0xff + 1 produces a carry */
	add;
	if carry cjmp transform_block_complete;
	jmp bad_block;

transform_block_complete:
	/* The transformation has been received and is correct. */
	/* Multiply it into the CTM_ADDR. */
	LDMAR CTM_ADDR;
	LDREG OP1 DRSRAM;
	LDMAR NCTM_ADDR;
	LDREG DST DRSRAM;
	LDREG OP2 MATRIX_B;
	LDMAR TMP1;
	LDSRAM 4;
	LDREG r4 transform_multiply_rtn;
	/* Assume TMP3-TMP5 are zero */
	jmp matrix_mult;
transform_multiply_rtn:
	/* swap CTM_ADDR and NCTM_ADDR */
	LDMAR CTM_ADDR;
	LDREG r1 DRSRAM;
	LDMAR NCTM_ADDR;
	LDREG r2 DRSRAM;
	LDSRAM DRREG r1;
	LDMAR CTM_ADDR;
	LDSRAM DRREG r2;
	jmp render_object;

/***********************************************************************
 * render_object -- multiply the object by the CTM, storing it in page
 * 2, and then give it to the video unit
 ***********************************************************************/
render_object:
	/* TMP3:OP1 = 1:VERT_LIST */
	LDREG OP1 VERT_LIST;	
	LDMAR TMP3;
	LDSRAM 1;

	/* TMP4:OP2 = 0:<CTM_ADDR> */
	/* assume TMP4 is already 0 */
	LDMAR CTM_ADDR;
	LDREG OP2 DRSRAM;

	/* TMP5:DST = 2:VERT_LIST */
	LDREG DST VERT_LIST;
	LDMAR TMP5;
	LDSRAM 2;

	/* TMP1 = <NUM_VERT> */
	LDMAR NUM_VERT;
	LDREG r4 DRSRAM;
	LDMAR TMP1;
	LDSRAM DRREG r4;
	
	LDREG r4 render_rtn;
	jmp matrix_mult;

render_rtn:
	jmp main_loop;

/***********************************************************************
 * object_block -- read in an object block and store it in memory.
 * Note that the vertex data in stored in page 1
 ***********************************************************************/
object_block:
	/* read in the number of vertices */
	LDMAR NUM_VERT;
	LDA 1;
	LDREG r4 object_read_vertices;
	jmp read_bytes;
object_read_vertices:
	/* read in the vertex data */
	LDA DRREG r1;
	LDMPR 1;
	LDMAR VERT_LIST;
	LDREG r4 object_read_vertices_1;
	jmp read_bytes;
object_read_vertices_1:
	LDA 1;
     	LDMPR 0;
	LDMAR TMP1;
	LDREG r4 object_read_vertices_2;
	jmp read_bytes;
object_read_vertices_2:
	add;
	if carry cjmp object_read_edges;
	jmp bad_block;

object_read_edges:
	/* read in the number of edges */
	LDMAR NUM_EDGE;
	LDA 1;
	LDREG r4 object_read_edges_1;
	jmp read_bytes;
object_read_edges_1:
	LDMAR EDGE_LIST;
	LDA DRREG r1;
	LDREG r4 check_eob_marker;
	jmp read_bytes;
	/* NOT REACHED */	
	jmp main_loop;

/***********************************************************************
 * read_bytes -- read a specified number of bytes into sequential SRAM
 * locations.
 *
 * Entry:
 * MAR contains starting address for block
 * A is the number of bytes to read
 * r4 is return addr
 *
 * Exit:
 * r1 contains last byte read
 * A = 0, B = 1
 * MAR is address after the block read
 ***********************************************************************/
read_bytes:
	LDB 1;

get_byte:
	if /dav_i cjmp get_byte;
	LDSRAM DRIN MARCNT;
	LDREG r1 DRIN;
	ACK_T;
get_byte_1:
	if /dav_i cjmp get_byte_2;
	jmp get_byte_1;
get_byte_2:
	ACK_T;

	LDA subc;
	if carry cjmp read_bytes_done;
	jmp get_byte;

read_bytes_done:
	jmp DRREG r4;

bad_block:
	jmp bad_block;

/*
 * The matrix multiply routine is stored in code segment 1.  The entry
 * point is at address 0xfb, which loads the cseg register and warps
 * to address 0x1fc, which then jumps to the routine at 0x100.  The
 * return is handled similarly, through a warp from 0x1fd to 0xfe.
 * 
 * matrix_mult is a cpage 0 entry point for this whole mess.
 */
matrix_mult:
	jmp matrix_mult_entry_0;

# SET_ADDRESS = fb;
matrix_mult_entry_0:
	LDCSEG 1; /* WARP! */
# SET_ADDRESS = fc;
# LOAD_ADDRESS = 1fc;
matrix_mult_entry_1:
	jmp matrix_mult_entry;

# SET_ADDRESS = fd;
# LOAD_ADDRESS = 1fd;
matrix_mult_return_1:
	LDCSEG 0; /* WARP! */
# SET_ADDRESS = fe;
matrix_mult_return_0:
	LDMAR TMP6;
	jmp DRSRAM;	

/***********************************************************************
 * Matrix multiply: <DST> <-- <OP1>*<OP2>
 * Entry:
 * TMP3:OP1, TMP4:OP2 are factors
 * TMP5:DST is result
 * TMP1 is the number of rows (vertices) in OP1
 * r4 is return addr
 *
 * Exit:
 * OP1, OP2, DST, TMP2, TMP6 are destroyed
 ***********************************************************************/
# SET_ADDRESS = 00;
# LOAD_ADDRESS = 100;
matrix_mult_entry:
	/* save return addr in TMP6 */
	LDMAR TMP6;
	LDSRAM DRREG R4;

do_row:
	/* inner loop: TMP2 is number of cols in OP2 (always 4) */
	LDMAR TMP2;
	LDSRAM 4;

	/* decrement and test outer loop counter */
	LDMAR TMP1;
	LDA DRSRAM;
	LDB 1;
	LDSRAM subc;
	if carry cjmp next_col;
	jmp mult_done;
	
next_col:
	/* debugging code */
	DRREG OP1;
	DRREG OP2;
	DRREG DST;

	/* decrement and test inner loop counter */
	LDMAR TMP2;
	LDA DRSRAM;
	LDB 1;
	LDSRAM subc;
	if carry cjmp inner_not_done;
	jmp next_row;

inner_not_done:
	/* set the current element to zero */
	LDMAR TMP5;
	LDMPR DRSRAM;
	LDMAR DST;
	LDSRAM 0;
	MARCNT;
	LDSRAM 0;
	LDMPR 0;

	LDREG r4 l1;
	jmp accum_prod;
l1:	LDREG r4 l2;
	jmp accum_prod;
l2:	LDREG r4 l3;
	jmp accum_prod;
l3:	LDREG r4 l4;
	jmp accum_prod;
l4:	

	/* update DST to next element */
	LDA DRREG DST;
	LDB 2;
	LDREG DST add;

	/* reset OP1 to beginning of row */
	LDA DRREG OP1;
	LDB 8;
	LDREG OP1 subc;
     
	/* update OP2 to use next column */
	LDA DRREG OP2;
	LDB 30;
	LDREG OP2 add;

	jmp next_col;

next_row:
	/* update OP1 to use next row */
	LDA DRREG OP1;
	LDB 8;
	LDREG OP1 add;

	/* reset OP2 to beginning of matrix */
	LDA DRREG OP2;
	LDB 8;
	LDREG OP2 subc;

	jmp do_row;

mult_done:
	jmp matrix_mult_return_1;

/*
 * accum_prod -- perform one multiplication-addition in a matrix
 * multiplication.  <DST> <-- <DST> + prod
 * 
 * Algorithm:
 * (1) load factor from OP1 into A and B, load into multiplier X
 * (2) load factor from OP2 into A and B, load into multiplier Y
 * (3) load lb of <DST> into A, lb of prod into B, add and
 * store result into lb of <DST> *and* latch CCs.
 * (4) repeat (3) with hb, adding a carry-in if lb addition produced a
 * carry out
 * (5) increment OP1 by 2, OP2 by 8
 *
 * Note: This "subroutine" assumes its return address is in r4
 */
accum_prod:
	/* A, B <- <OP1> */
	LDMAR TMP3;
	LDMPR DRSRAM;
	LDMAR DRREG OP1;
	LDA DRSRAM;
	MARCNT;
	LDB DRSRAM;
	LDMPR 0;
	LDX;

	/* A, B <- <OP2> */
	LDMAR TMP4;
	LDMPR DRSRAM;
	LDMAR DRREG OP2;
	LDA DRSRAM;
	MARCNT;
	LDB DRSRAM;
	LDMPR 0;
	LDY;

	/* set all sram references to DST page */
	LDMAR TMP5;
	LDMPR DRSRAM;

	/* <dst+1> <- <dst+1> + low prod */
	LDMAR DST;
	MARCNT;
	LDA DRSRAM;
	LDB DRPR_LO;
	LDSRAM add;

	/* <dst> <- <dst> + hi prod + carry */
	LDMAR DST;
	LDA DRSRAM;
	LDB DRPR_HI;
	if carry cjmp accum_with_carry;
	LDSRAM add;
	jmp done_with_accum;
accum_with_carry:
	LDSRAM addc;

done_with_accum:
	LDMPR 0;

	/* OP1 += 2, OP2 += 8 */
	LDA DRREG OP1;
	LDB 2;
	LDREG OP1 add;

	LDA DRREG OP2;
	LDB 8;
	LDREG OP2 add;
     
	jmp DRREG r4;
