/*
 * Instruction word:   
 *
 *      conditional jumps               ccc..ddd ....mmmm aaaaaaaa 
 * 	unconditional jumps		001..ddd ....mmmm aaaaaaaa
 * 
 *      assertions                      000rrddd llllmmmm .....ttt
 *      immediate value                 000rr000 llll.... vvvvvvvv
 *
 * where c = condition code select
 * 	 m = ALU mode
 *       d = drive signals
 *       a = address or unused, see below
 *       l = load signals
 * 	 r = register read selection
 *       R = drive debug rom
 *       t = control signals
 *       v = immediate value
 *       . = not connected/ignored
 * 
 * In a jump instruction, the ccc bits select which condition code to
 * conditional jump on (condition 1 is wired always to jump).  The new
 * address comes from the device selected by ddd; the default (000,
 * used when nothing else is specified) is the low ROM.  If some other
 * device is selected, the low ROM bits are ignored (and will probably
 * be zero).  The mmmm bits control the ALU in case ddd selects DRALU.
 * 
 * In an assertion statement, the ccc bits select condition 0 (the
 * default) which is wired never to jump.  All control bits have their
 * labeled meanings.
 *
 * An immediate value statement is an assertion statement in which
 * DRROM is selected by ddd.  The low ROM bits are interpreted as a
 * value, so no register selection or control signals are possible.
 * The llll bits control where the value is loaded.
 */

op <23:0>;
address op <7:0>;
value op <7:0>;

/*
 * Uncondition instructions
 *
 * The selected CC for these instructions are wired to Vcc or GND
 * according to whether the instruction is a jump or not.
 */ 
assert	op<23:21>=%b000;	/* standard assertion statement */
jmp	op<23:21>=%b001;	/* jump to addr in low rom */
brk	op<23:21>=%b111;	/* halt the simulator */
noop	op<23:0>=0;		/* do nothing */

/* 
 * Conditional instructions (jumps)
 *
 * The selected CC is passed to the /LD input of the program counter;
 * therefore, all conditions are active-low.  The condition names are
 * choosen to reflect how they behave for the programmer, and not
 * their actual polarity.
 */
carry	op<23:21>=%b010;
/equal	op<23:21>=%b011;
/p0	op<23:21>=%b100;
/ack_i	op<23:21>=%b101;
/dav_i	op<23:21>=%b110;

if	nop;
cjmp	nop;

/*
 * Drive selection, default = DRROM
 */
DRROM	op<18:16>=%h0;
DRALU	op<18:16>=%h1;
DRREG	op<18:16>=%h2;
DRPR_LO	op<18:16>=%h3;
DRPR_HI	op<18:16>=%h4;
DRSRAM	op<18:16>=%h5;
DRSWT	op<18:16>=%h6;
DRIN   	op<18:16>=%h7;

/*
 * A "do" statement means that "nothing" is being driven on the bus.
 * It is therefore safe to drive anything on the bus.  Anything is
 * acceptable except for DRROM, since that causes the control PAL to
 * interpret the data as an immediate value.
 */
do	DRALU;

/*
 * Load selection, default = load nothing
 */
LDA	op<15:12>=%h1;
LDB	op<15:12>=%h2;
LDX	op<15:12>=%h3;
LDY	op<15:12>=%h4;
LDREG	op<15:12>=%h5;
LDSRAM	op<15:12>=%h6;
LDMAR	op<15:12>=%h7;
LDMPR	op<15:12>=%h8;
LDOUT	op<15:12>=%h9;
LDCSEG	op<15:12>=%hA;

/*
 * Register selection
 */
r1	op<20:19>=%h0;
r2	op<20:19>=%h1;
r3	op<20:19>=%h2;
r4	op<20:19>=%h3;

/*
 * Control selection
 */
MARCNT	op<2:0>=%h1;
Cin	op<2:0>=%h2;
DRCONST op<2:0>=%h3;

CLR_T	op<2:0>=%h4;
ACK_T	op<2:0>=%h5;
DAV_T	op<2:0>=%h6;

/*
 * ALU operation controls, default = A
 *
 * The ALU is wired into arithmetic mode.  /Cin is a control signal.
 */
a	DRALU op<11:8>=%h0;
aplus1	DRALU op<11:8>=%h0 Cin;
sub	DRALU op<11:8>=%h6;
subc	DRALU op<11:8>=%h6 Cin;
add	DRALU op<11:8>=%h9;
addc	DRALU op<11:8>=%h9 Cin;
latch	sub;

/*
 * Convenient symbolic names.  See the memory map in math.as.
 */
OP1 r1;
OP2 r2;
DST r3;

MATRIX_A op<7:0> = 0;
MATRIX_B op<7:0> = 32;
MATRIX_C op<7:0> = 64;
CTM_ADDR op<7:0> = 96;
NCTM_ADDR op<7:0> = 97;
TMP1 op<7:0> = 98;
TMP2 op<7:0> = 99;
TMP3 op<7:0> = 100;
TMP4 op<7:0> = 101;
TMP5 op<7:0> = 102;
TMP6 op<7:0> = 103;
TMP7 op<7:0> = 104;
TMP8 op<7:0> = 105;
EYE_POINT op<7:0> = 106;
FIX_POINT op<7:0> = 112;
NUM_VERT op<7:0> = 118;
NUM_EDGE op<7:0> = 119;
EDGE_LIST op<7:0> = 120;
VERT_LIST op<7:0> = 0;

OBJECT op<7:0> = 1;
EYE op<7:0> = 2;
FIX op<7:0> = 3;
IDENTIFY op<7:0> = 4;
TRANSFORM op<7:0> = 5;


