#include <stdio.h>
#include "Matrix.h"
#include "CohSuthClip.h"
extern "C" {
#include <tgif/tgif.h>
}

void CohSuthClipDrawLine(double, double, double, double, double,
			 double, double, double);
static double xmin, xmax, ymin, ymax;
static int debug = 0;

void CohSuthSetDebug(int d)
{
     debug = d;
     if (debug) printf("CohSuth debugging on.\n");
}

void CohSuthSetClip(double x, double y, double X, double Y)
{
     xmin = x; ymin = y; xmax = X; ymax = Y;
     if (debug)
	  printf("CohSuth: clipping to (%6.3f, %6.3f),(%6.3f, %6.3f)\n",
		 x, y, X, Y);
}

inline char Outcode(double x, double y, double xmin, double ymin,
		    double xmax, double ymax) 
   // Returns the outcode of (x,y) in the rectangle (xmin,ymin)-(xmax,ymax).
   // Bits are TBLR.
{
     char c = 0;
     
     if (y > ymax)
	  c |= 0x08;
     if (y < ymin)
	  c |= 0x04;
     if (x < xmin)
	  c |= 0x02;
     if (x > xmax)
	  c |= 0x01;

     // This should work:
     // char c = (y>ymax)<<3 + (y<ymin)<<2 + (x<xmin)<<1 + (x>xmax);
     
     return c;
}

inline int Top(char c)
{
     return c & 0x8;
}

inline int Bottom(char c)
{
     return c & 0x4;
}

inline int Left(char c)
{
     return c & 0x2;
}

inline int Right(char c)
{
     return c & 0x1;
}
	
void CohSuthClipDraw(Matrix &m)
   // Draws Matrix m clipped to the current clipping region using the
   // Cohen-Sutherland clipping algorithm.
{
     int i;

     for (i = 1; i < m.Rows(); i += 1)
	  CohSuthClipDrawLine(m(i, 1), m(i, 2), m(i+1, 1), m(i+1, 2),
			      xmin, ymin, xmax, ymax);
     CohSuthClipDrawLine(m(m.Rows(), 1), m(m.Rows(), 2), m(1, 1), m(1, 2),
			 xmin, ymin, xmax, ymax);
}

void CohSuthClipDrawLine(double x0, double y0, double x1, double y1,
			 double xmin, double ymin, double xmax,
			 double ymax) 
   // Clips the line (x0,y0)-(x1,y1) to the rectangle (xmin,ymin),
   // (xmax,ymax) using the Cohen-Sutherland algorithm and draws it.
{
     char out0 = Outcode(x0, y0, xmin, ymin, xmax, ymax);
     char out1 = Outcode(x1, y1, xmin, ymin, xmax, ymax);
     char out;
     double x, y, xd, yd;

     if (debug)
	  printf("CohSuth: line (%6.3f, %6.3f)-(%6.3f, %6.3f)\n", x0,
		 y0, x1, y1);
     
     while (1) {
	  
	  /* Trivally accept */
	  if (! out0 && ! out1) {
	       if (debug)
		    printf("CohSuth: accept\n");
	       break;
	  }

	  /* Trivially reject */
	  if (out0 & out1) {
	       if (debug)
		    printf("CohSuth: reject\n");
	       return;
	  }

	  if (out0) {
	       out = out0;
	       xd = x0; yd = y0;
	  } else {
	       out = out1;
	       xd = x1; yd = y1;
	  }

	  /* At least one of the bits must be set */
	  if (Top(out)) {
	       x = x0 + (x1-x0) * (ymax-y0) / (y1-y0);
	       y = ymax;
	       if (debug)
		    printf("CohSuth: clipping (%6.3f, %6.3f) to "
			   "(%6.3f, %6.3f)\n", xd, yd, x, y);
	  } else if (Bottom(out)) {
	       x = x0 + (x1-x0) * (ymin-y0) / (y1-y0);
	       y = ymin;
	       if (debug)
		    printf("CohSuth: clipping (%6.3f, %6.3f) to "
			   "(%6.3f, %6.3f)\n", xd, yd, x, y);
	  } else if (Left(out)) {
	       y = y0 + (y1-y0) * (xmin-x0) / (x1-x0);
	       x = xmin;
	       if (debug)
		    printf("CohSuth: clipping (%6.3f, %6.3f) to "
			   "(%6.3f, %6.3f)\n", xd, yd, x, y);
	  } else /* if (Right(out)) */ {
	       y = y0 + (y1-y0) * (xmax-x0) / (x1-x0);
	       x = xmax;
	       if (debug)
		    printf("CohSuth: clipping (%6.3f, %6.3f) to "
			   "(%6.3f, %6.3f)\n", xd, yd, x, y);
	  }

	  /* Move clipped point and iterate */
	  if (out == out0) {
	       x0 = x;
	       y0 = y;
	       out0 = Outcode(x0, y0, xmin, ymin, xmax, ymax);
	  }
	  else {
	       x1 = x;
	       y1 = y;
	       out1 = Outcode(x1, y1, xmin, ymin, xmax, ymax);
	  }    
     }

     Line(int(x0), int(y0), int(x1), int(y1));
}
