#include <stdio.h>
#include <stdlib.h>
#include "Matrix.h"
#include "MatrixUtil.h"
#include "Hedron.h"
#include "transform.h"
#include "Clipping.h"

const int WIDTH = 300;
const int HEIGHT = 300;

extern "C" {
#include <tgif/tgif.h>
}

int debug = 0;

void PrintMatrix(char *msg, Matrix &m)
{
     printf("%s\n", msg);
     m.Print();
}

Hedron ClipHedron(Hedron& h, double xmin, double xmax, double ymin,
		  double ymax, double hither, double yon)
{
     Matrix clipper(2, 3,
		    xmin, ymin, hither,
		    xmax, ymax, yon);

     PrintMatrix("clipper", clipper);

     return h.Clip(clipper);
}


void DrawPolyhedron(Hedron& h)
{
     int i;

     for (i = 0; i < h.NumEdges(); i += 1)
          Line(int(h.VertexCoord(2*i, 1)),
               int(h.VertexCoord(2*i, 2)),
               int(h.VertexCoord(2*i+1, 1)),
               int(h.VertexCoord(2*i+1, 2)));
}

Matrix PreClipTransform(Matrix& Peye_w, Matrix& Pfix_w, Matrix& UP_w)
{
     Matrix xhat(1, 3), yhat(1, 3), zhat(1, 3);
     Matrix View(1, 3);
     View = Peye_w - Pfix_w;

     Matrix ViewHat(Unitize(View));
     xhat = Unitize(CrossProd(UP_w, ViewHat));
     yhat = CrossProd(ViewHat, xhat);
     zhat = -1.0*ViewHat;

     Matrix Rot(4, 4,
		xhat(1, 1), yhat(1, 1), zhat(1, 1), 0.0, 
		xhat(1, 2), yhat(1, 2), zhat(1, 2), 0.0, 
		xhat(1, 3), yhat(1, 3), zhat(1, 3), 0.0, 
		0.0,        0.0,        0.0,        1.0);
     
     Matrix V(Translator(-Peye_w(1, 1), -Peye_w(1, 2), -Peye_w(1, 3)) * Rot);
     if (debug) {
	  PrintMatrix("View:", View);
	  PrintMatrix("Rotation matrix:", Rot);
	  PrintMatrix("V:", V);
     }

     return V;
}

void ParseParamFile(char *filename, Matrix *Peye, Matrix *UP, Matrix
		    *Pfix, Hedron **h, double *S, double *D, double *F)
   // Fills in all arguments after the first.
{
     FILE *f;
     double a, b, c;
     int i, j, k, l;

     f = fopen(filename, "r");
     if (f == NULL) {
	  fprintf(stderr, "Cannot open parameter file %s.\n", filename);
	  exit(1);
     }

     /* Perspective parameters */
     fscanf(f, "%f %f %f", S, D, F);
     if (debug)
	  printf("Perspective:\t S=%6.3f D=%6.3f F=%6.3f\n", *S, *D, *F);

     /* Eye point */
     fscanf(f, "%f %f %f", &a, &b, &c);
     (*Peye)(1, 1) = a; (*Peye)(1, 2) = b; (*Peye)(1, 3) = c;
     if (debug)
	  printf("Eye point:\t %6.3f %6.3f %6.3f\n", a, b, c);

     /* UP vector */
     fscanf(f, "%f %f %f", &a, &b, &c);
     (*UP)(1, 1) = a; (*UP)(1, 2) = b; (*UP)(1, 3) = c;
     if (debug)
	  printf("UP vector:\t %6.3f %6.3f %6.3f\n", a, b, c);

     /* Fixation point */
     fscanf(f, "%f %f %f", &a, &b, &c);
     (*Pfix)(1, 1) = a; (*Pfix)(1, 2) = b; (*Pfix)(1, 3) = c;
     if (debug)
	  printf("Fixation point:\t %6.3f %6.3f %6.3f\n", a, b, c);

     /* Hedron matrix */
     fscanf(f, "%d", &i);
     if (debug)
	  printf("Hedron has %d vertices.\n", i);
     
     Matrix *m = new Matrix(i, 4);
     for (j = 1; j <= i; j ++) {
	  fscanf(f, "%f %f %f", &a, &b, &c);
	  (*m)(j, 1) = a; (*m)(j, 2) = b; (*m)(j, 3) = c; (*m)(j, 4) = 0.0;
	  if (debug)
	       printf("Vertex:\t\t %6.3f %6.3f %6.3f\n", a, b, c);
     }

     /* Hedron edges */
     fscanf(f, "%d", &i);
     if (debug)
	  printf("Hedron has %d edges.\n", i);
     
     *h = new Hedron(*m, i);
     for (j = 0; j < i; j++) {
	  fscanf(f, "%d %d", &k, &l);
	  h->Edges()[2*j] = k;
	  h->Edges()[2*j+1] = l;
	  if (debug)
	       printf("Edge:\t\t %d %d\n", k, l);
     }

     delete m;
     fclose(f);
}

Hedron Perspective(const Hedron& h1, const double d)
{
     Hedron h2(h1);
     double v;
     int i;

     for (i = 1; i <= h2.Vertices().Rows(); i++) {
	  if (h2.Vertices()(i, 3) == 0.0) {
	       h2.Vertices()(i, 1) = HUGE;
	       h2.Vertices()(i, 2) = HUGE;
	  }
	  else {
	       v = h2.Vertices()(i, 3) / d;
	       h2.Vertices()(i, 1) /= v;
	       h2.Vertices()(i, 2) /= v;
	  }
     }

     return h2;
}

void problem1(char *filename)
{
     Matrix CTM(4, 4), Peye(1, 3), UP(1, 3), Pfix(1, 3), V(4, 4), *points;
     Hedron *h;
     double D, S, F, d_s, d_f, border = 20.0, hither = 10.0, yon = 500.0;
     double a, b, c;
     char z;

     /* Get the viewing situation */
     ParseParamFile(filename, &Peye, &UP, &Pfix, &h, &S, &D, &F);

     // Hardcoded because of some weird pointer bug
     points = new Matrix(8,4,
		       -50,-50,-50,1,
		       -50,50,-50,1,
		       50,50,-50,1,
		       50,-50,-50,1,
		       -50,-50,50,1,
		       -50,50,50,1,
		       50,50,50,1,
		       50,-50,50,1);

     h->Vertices() = *points;
     
     int edges[] = {1,2, 2,3, 3,4, 4,1,
		  5,6, 6,7, 7,8, 8,5,
		  1,5, 2,6, 3,7, 4,8};
     h = new Hedron(*points,12,edges);

     CTM.Identify();

     while (1) {
	  Clear();

	  double PerMag = MatrixMag(Peye - Pfix);
	  V = PreClipTransform(Peye, Pfix, UP);

#if 0
	  DrawPolyhedron(Perspective((*h) * CTM * V, PerMag) *
			 Translator(150, 150, 0));
#endif 0
	  Hedron world((*h) * CTM);
	  Hedron eye(world*V);
	  Hedron perspect(Perspective(eye, PerMag));
	  Hedron clipped(ClipHedron(perspect,
				    -(WIDTH/2.0-border),WIDTH/2.0-border,
				    -(HEIGHT/2.0-border),HEIGHT/2.0-border,
				    hither, yon));

	  DrawPolyhedron(clipped * Translator(150, 150, 0));
	  ForceOutput();

	  if (scanf("%1s", &z) == EOF)
	       break;

	  switch (z) {
	  case 'c':
	  case 'C':
	       scanf("%f %f %f", &a, &b, &c);
	       border = a; hither = b; yon = c;
	       break;
	  case 'e':
	  case 'E':
	       scanf("%f %f %f", &a, &b, &c);
	       Peye(1, 1) = a; Peye(1, 2) = b; Peye(1, 3) = c;
	       break;
	  case 'f':
	  case 'F':
	       scanf("%f %f %f", &a, &b, &c);
	       Pfix(1, 1) = a; Pfix(1, 2) = b; Pfix(1, 3) = c;
	       break;
	  case 'u':
	  case 'U':
	       scanf("%f %f %f", &a, &b, &c);
	       UP(1, 1) = a; UP(1, 2) = b; UP(1, 3) = c;
	       break;
	  case 't':
	  case 'T':
	       scanf("%f %f %f", &a, &b, &c);
	       CTM = CTM * Translator(a, b, c);
	       break;
	  case 's':
	  case 'S':
	       scanf("%f %f %f", &a, &b, &c);
	       CTM = CTM * Scaler(a, b, c);
	       break;
	  case 'r':
	  case 'R':
	       scanf("%1s %f", &z, &a);
	       switch (z) {
	       case 'X':
	       case 'x':
		    CTM = CTM * Rotator_X(a);
		    break;
	       case 'Y':
	       case 'y':
		    CTM = CTM * Rotator_Y(a);
		    break;
	       case 'Z':
	       case 'z':
		    CTM = CTM * Rotator_Z(a);
		    break;
	       }
	       break;
	  case 'i':
	  case 'I':
	       CTM.Identify();
	       break;
	  }
     }

     /* Reset the EOF condition */
     clearerr(stdin);
     printf("\n");

     delete h;
}
     
main(int argc, char **argv)
{
     Panel p;

     if (argc < 2) {
	  fprintf(stderr, "Usage: ps4 filename.\n");
	  exit(1);
     }

     p = NewPanel(WIDTH, HEIGHT);
     SetFunction(DRAW);
     UsePanel(p);

     problem1(argv[1]);

     DoneWithPanel(p);
     return 0;
}
