;;; -*- Mode: LISP; Syntax: Common-lisp; Package: USER; Base: 10; -*-

(in-package 'USER)

;;; The clauses for the example.  The negated goal is the last one.

(setq block-test-clauses
      '((1 (x y s) 
	   ((not clear x s) (not clear y s) (equal x y) (on x y (puton x y s))))
        (2 (x y z s)
	   ((not on x y s) (not clear x s) (not clear z s) (equal y z)
	    (equal x z) (clear y (puton x z s))))
        (3 (x y s u)
	   ((not clear u s) (equal u y) (clear u (puton x y s))))
        (4 (s) ((clear table s)))
        (5 () ((on b a s0)))
        (6 () ((on a table s0)))
        (7 () ((clear b s0)))
        (8 () ((clear c s0)))
        (9 () ((on c table s0)))
	;;; Some of the "trivial" axioms ...
	(11 () ((not equal a table)))
	(12 () ((not equal b table)))
	(13 () ((not equal c table)))
        (14 () ((not equal a b)))
        (15 () ((not equal a c)))
        (16 () ((not equal b a)))
        (17 () ((not equal b c)))
        (18 () ((not equal c a)))
        (19 () ((not equal c b)))
	;;; The negated goal...
	;;; Placed at the end for ease of access.
        (10 (s) ((not clear a s) (not clear c s)))))

;;; If you do (EVAL BLOCK-TEST), it will churn away.
;;; WARNING: get as FAST a machine as you can ...

(setq block-test
      '(tpu (butlast block-test-clauses) (last block-test-clauses)))

;;; BASIC DEFINITIONS AND ACCESS FUNCTIONS

(defparameter *MAX-CLAUSE-NUMBER* 0)

(defparameter *PRINT-RESOLVENTS* NIL)	; Prints trace of resolutions

;;; Variable name lists used to standardize clauses apart.  That is
;;; the variables in clauses to be resolved are renamed to members of
;;; this list so that there are no common variables in the clauses.

(defparameter XLIST '(x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13 x14 x15))
(defparameter YLIST '(y1 y2 y3 y4 y5 y6 y7 y8 y9 y10 y11 y12 y13 y14 y15))
(defparameter ZLIST '(zz1 zz2 zz3 zz4 zz5 zz6 zz7 zz8 zz9 zz10 zz11
		      zz12 zz13 zz14 zz15))

;;; The representation of a clause is:
;;; (<clause number> <list of vars> <list of literals>)
;;; These functions access that representation.

(defun CLAUSE-NUMBER (clause) (first clause))
(defun CLAUSE-VARS (clause) (second clause))
(defun CLAUSE-LITERALS (clause) (third clause))

;;; A literal is either:
;;; Positive: (<predicate> . <list of args>)
;;; Negative: (NOT <predicate> . <list of args>)

(defun NEGATIVE (literal) (eq (first literal) 'NOT))
(defun POSITIVE (literal) (not (eq (first literal) 'NOT)))

(defun VARIABLE? (x) 
  (and (atom x)
       (or (member x xlist :test #'eq)
	   (member x ylist :test #'eq))))

;;; TOP LEVEL PROCEDURE.
;;; Returns (CONTRADICTION clause-number clause-number)
;;;         List of records of resolutions of the form 
;;;           (New-clause-number resolvent-number-1 resolvent-number-2
;;;           literal-#-in-1 literal-#-in-2)
;;;         List of all the clauses not in the input  that were ever
;;;          used in a resolution step.
;;; or returns FAIL.

(defun TPU (input-clauses waiting-to-resolve &optional all-clauses)
  (setq *max-clause-number*		; Used in assigning new clause numbers.
	(max-clause-number (append input-clauses waiting-to-resolve)))
  (tpu-aux input-clauses waiting-to-resolve all-clauses nil))

(defun TPU-AUX (input-clauses waiting-to-resolve all-clauses history
			      &optional (cycles 0))
  (format t "~% CYCLE=~s, ~s clauses waiting to resolve ..." 
	  cycles (length waiting-to-resolve))
  (if (null waiting-to-resolve)
      'fail
      (progn
	;; Don't resolve with a clause that has previously been used.
	(dolist (x waiting-to-resolve)
	  (if (member (clause-literals x) all-clauses
		      :test #'equal :key #'clause-literals)
	      ;; The top of the sorted queue is in ALL-CLAUSES, that is, has
	      ;; been used before, so remove it.
	      (setq waiting-to-resolve (rest waiting-to-resolve))
	      ;; A previously unseen clause, so go on.
	      (return t)))
	(multiple-value-bind (resolvents new-history)
	    ;; Resolve the "best" clause against all the input clauses
	    ;; Note that it only resolves against INPUT clauses here,
	    ;; that is why this is called INPUT resolution...
	    (resolve* (first waiting-to-resolve) input-clauses)
	  ;; This test is just for safety.
	  (if (> cycles 500)
	      (if (y-or-n-p "Do you wish to continue? ")
		  (setq cycles 0.)
		  (error "This has gone on long enough ...")))
	  (if (eq (first resolvents) 'contradiction)
	      ;; Found a contradiction, that is, proved the goal.
	      (values resolvents
		      (append new-history history)
		      (cons (first waiting-to-resolve) all-clauses))
	      ;; Keep cursing ...
	      (tpu-aux input-clauses
		       (merge-new-clauses resolvents waiting-to-resolve)
		       (cons (first waiting-to-resolve) all-clauses)
		       (append new-history history)
		       (1+ cycles)))))))

;;; A simple comparison function based on number of variables in the clause

(defun list-depth (l)
  (list-depth-aux l 1))

(defun list-depth-aux (l d)
  (cond ((null l) d)
	((atom (car l)) (list-depth-aux (cdr l) d))
	(t (max (list-depth-aux (car l) (1+ d))
		(list-depth-aux (cdr l) d)))
	))

(defun CLAUSE-IS-BETTER (clause-1 clause-2)
  (let ((len1 (length (clause-vars clause-1)))
	(len2 (length (clause-vars clause-2))))
    (if (not (eq len1 len2))
	(< len1 len2)
	(> (list-depth (clause-literals clause-1))
	   (list-depth (clause-literals clause-1))))
    ))


;;(defun CLAUSE-IS-BETTER (c1 c2)
;;  (let ((ld1 (list-depth (clause-literals c1)))
;;	(ld2 (list-depth (clause-literals c2))))
;;    (if (not (eq ld1 ld2))
;;	(> ld1 ld2)
;;	(< (length (clause-vars c1)) (length (clause-vars c2))))
;;   ))

;;(defun CLAUSE-IS-BETTER (c1 c2)
;;  (> (clause-number c1) (clause-number c2)))

;;(defun CLAUSE-IS-BETTER (clause-1 clause-2)
;;  (< (length (clause-vars clause-1)) (length (clause-vars clause-2))))

(defun COUNTATOMS (l &optional (count 0))
  (cond ((null l) count)
	((atom l) (1+ count))
	((listp l)
	 (+ (countatoms (first l)) (countatoms (rest l))))))

(defun MERGE-NEW-CLAUSES (new-clauses sorted-old-clauses)
  (let ((sorted-new-clauses (sort (copy-list new-clauses) #'clause-is-better)))
    (cond ((null sorted-new-clauses) sorted-old-clauses)
	  ((null sorted-old-clauses) sorted-new-clauses)
	  (t
	   ;;; This does a merge of two sorted lists to produce a combined
	   ;;; sorted list.  This is to avoid resorting the whole queue.
	   (merge 'list
		  sorted-new-clauses
		  (copy-list sorted-old-clauses)
		  #'clause-is-better)))))

;;; UNIFICATION algorithm

(defun UNIFICATION (e1 e2)
  "Returns the most general unifier of two expressions."
  (if (not (= (length e1) (length e2)))
    'no
    (unification-aux e1 e2 nil)))

(defun UNIFICATION-AUX (e1 e2 unifier)
  (let ((disagreement (disagree e1 e2)))
    (if (null disagreement)
      (reverse unifier)
      (let ((d1 (first disagreement))
            (d2 (second disagreement)))
      (cond ((variable? d1)
             (if (inside d1 d2)
               'no
               (unification-aux (subst d2 d1 e1)
                                (subst d2 d1 e2)
                                (cons disagreement unifier))))
            ((variable? d2)
             (if (inside d2 d1)
               'no
               (unification-aux (subst d1 d2 e1)
                                (subst d1 d2 e2)
                                (cons (reverse disagreement) unifier))))
            (t 'no))))))

;;; First rename all the variables using ZLIST, then rename them using
;;; VAR-LIST.  This is to avoid introducing name conflicts from
;;; previous times when we may have renamed variables in clause using
;;; VAR-LIST.

(defun RENAME (clause var-list)
  "Rename the variables in CLAUSE using the variable names in VAR-LIST"
  (let ((new-clause (rename-aux (clause-vars clause) clause zlist)))
    (rename-aux (clause-vars new-clause) new-clause var-list)))

(defun RENAME-AUX (vars clause new-vars)
  (if (null vars) 
    clause
    (rename-aux (rest vars)
                (subst (first new-vars) (first vars) clause)
                (rest new-vars))))

(defun INSIDE (a e)
  "Tests whether an atom A appears in expression E"
  (cond ((atom e) (eq a e))
        ((inside a (first e)) t)
        (t (inside a (rest e)))))

(defun DISAGREE (e1 e2)
  "Returns a list describing where two lists (of equal length) disagree."
  (cond ((null e1) nil)
        ((or (atom e1) (atom e2))
         (cond ((equal e1 e2) nil)
               (t (list e1 e2))))
        ((equal (first e1) (first e2))
         (disagree (rest e1) (rest e2)))
        ((or (atom (first e1)) (atom (first e2)))
         (list (first e1) (first e2)))
        (t (disagree (first e1) (first e2)))))

;;; RESOLUTION

(defun RESOLVE* (clause list-of-clauses)
  (if (null list-of-clauses)
      nil
      (let ((resolvent (resolve clause (first list-of-clauses))))
	(if (eq (first resolvent) 'contradiction) 
	    (values resolvent nil)
	    (multiple-value-bind (resolvents history)
		(resolve* clause (rest list-of-clauses))
	      (values (append (first resolvent) resolvents)
		      (append (second resolvent) history)))))))

(defun RESOLVE (c1-input c2-input)
  (let ((c1 (rename c1-input xlist))
        (c2 (rename c2-input ylist)))
    (let ((result (resolve-aux c1 c2 (clause-literals c1) (clause-literals c2) nil)))
      (unless (or (null *print-resolvents*) (null (first result)))
	
	(format t "~% RESOLVE called with:")
	(format t "~%   ~s " c1-input)
	(format t "~%   ~s " c2-input)
	(format t "~% Resolvents are:")
	(print-resolvents result))
      result)))	

(defun RESOLVE-AUX (c1 c2 c1-lits c2-lits prev-c1-lits &optional (c1-index 0))
  (if (null c1-lits)
      (list nil nil)
      (let ((resolvent1 (uresolve c1 c2
				  c1-lits c2-lits
				  prev-c1-lits nil
				  c1-index 0))
	    (resolvent-rest (resolve-aux c1 c2 (rest c1-lits) c2-lits
					 (cons (first c1-lits)
					       prev-c1-lits) (1+ c1-index))))
	(if (eq (first resolvent1) 'contradiction)
	    resolvent1
	    (mapcar #'append resolvent1 resolvent-rest)))))

(defun URESOLVE (c1 c2 c1-lits c2-lits prev-c1-lits prev-c2-lits
		    c1-index c2-index 
		    &optional
		    resolvents history)
  "Resolve a single literal from c1 with literals from c2."
  (if (null c2-lits)
      (list (reverse resolvents) (reverse history))
      (let ((unifier (unify-possible-resolvents (first c1-lits) (first c2-lits))))
	(if (eq unifier 'no)
	    ;; Did not unify, so keep looking
	    (uresolve c1 c2
		      c1-lits (rest c2-lits)
		      prev-c1-lits (cons (first c2-lits) prev-c2-lits)
		      c1-index (1+ c2-index)
		      resolvents history)
	    (if (and (= c2-index 0) (= c1-index 0) 
		     (null (rest c1-lits)) (null (rest c2-lits)))
		;; We found a resolution for unit clauses, so we are
		;; done.
		(list 'CONTRADICTION (clause-number c1) (clause-number c2))
		(let* ((r (append (reverse prev-c1-lits) (rest c1-lits) 
				 (reverse prev-c2-lits) (rest
							 c2-lits)))
		       (new-literals (instantiate r unifier)))
		  (uresolve c1 c2
			    c1-lits (rest c2-lits)
			    prev-c1-lits (cons (first c2-lits) prev-c2-lits)
			    c1-index (1+ c2-index)
			    (cons (list (incf *max-clause-number*)
					(variables-of new-literals)
					new-literals)
				  resolvents)
			    (cons (list *max-clause-number*
					(clause-number c1) (clause-number c2)
					c1-index c2-index)
				  history))))))))

(defun UNIFY-POSSIBLE-RESOLVENTS (literal-1 literal-2)
  (cond ((and (positive literal-1) (negative literal-2))
         (unification literal-1 (rest literal-2)))
        ((and (negative literal-1) (positive literal-2))
         (unification (rest literal-1) literal-2))
        (t 'no)))

(defun UNIFY? (l1 l2) (not (eq 'no (unification l1 l2))))

(defun VARIABLES-OF (x &optional variables)
  (cond ((null x) variables)
	((atom x)
	 (if (variable? x)
	     (adjoin x variables)
	     variables))
	((listp x)
	 (variables-of (rest x)
		       (variables-of (first x) variables)))))

(defun INSTANTIATE (literals unifier)
  (if (or (null literals) (null unifier))
      literals
      (let ((var-value (first unifier)))
	(instantiate (subst (second var-value) (first var-value) literals)
		     (rest unifier)))))

(defun MAX-CLAUSE-NUMBER (clauses)
  (let ((max -1))
    (dolist (clause clauses)
      (let ((clause-number (clause-number clause)))
        (if (> clause-number max)
	    (setq max clause-number))))
    max))

(defun PRINT-RESOLVENTS (results)
  (if (eq (first results) 'contradiction)
      (print results)
      (dolist (r (first results)) (print r))))


