#include <math.h>
#include "MatrixMisc.H"
#include "LightSolid.H"

inline double DotProd3(Matrix& m1, Matrix& m2)
{
     return m1(1, 1)*m2(1, 1)+m1(1, 2)*m2(1, 2)+m1(1, 3)*m2(1, 3);
}

Solid ConstLightSolid(Solid& solid, double Ia, Matrix& lsource, Matrix&
		       viewpoint, double Ip, double n)
{
     Matrix N(1, 3), P(1, 3), L(1, 3), R(1, 3), V(1, 3);
     double Iaka, Ipkd, Ipks, NdotL, RdotV, bright;

     /* For each face in the solid, */
     for (int fn = 0; fn < solid.NumFaces(); fn++) {
	  SolidFace *face = solid.Face(fn);

	  Iaka = face->ka*Ia;
	  Ipkd = face->kd*Ip;
	  Ipks = face->ks*Ip;
	  
	  int vi = face->VIndxs()[0];
	  int ni = face->NIndxs()[0];

	  /* P = current vertex point = first point in the polygon */
	  P(1, 1) = solid.Vertices()(vi, 1);
	  P(1, 2) = solid.Vertices()(vi, 2);
	  P(1, 3) = solid.Vertices()(vi, 3);

	  /* N = normal at the current vertex = only normal available */
	  N(1, 1) = solid.Normals()(ni, 1);
	  N(1, 2) = solid.Normals()(ni, 2);
	  N(1, 3) = solid.Normals()(ni, 3);

	  L = Unitize(lsource - P);
	  V = Unitize(viewpoint - P);

	  NdotL = DotProd3(N, L);
	  R = 2.0*NdotL*N - L;
	  RdotV = DotProd3(R, V);

	  if (NdotL < 0.0) 
	       NdotL = 0.0;
		    
	  if (RdotV < 0.0)
	       RdotV = 0.0;

	  face->Colors()[0] = Iaka + Ipkd*NdotL + Ipks*pow(RdotV, n);
     }

     return solid;
}

Solid GouraudLightSolid(Solid& solid, double Ia, Matrix& lsource, Matrix&
			 viewpoint, double Ip, double n)
{
     Matrix N(1, 3), P(1, 3), L(1, 3), R(1, 3), V(1, 3);
     double Iaka, Ipkd, Ipks, NdotL, RdotV, bright;

     /* For each face in the solid, */
     for (int fn = 0; fn < solid.NumFaces(); fn++) {
	  SolidFace *face = solid.Face(fn);

	  /* For each polygon vertex (of the face), */
	  for (int pv = 0; pv < face->NumVerts(); pv++) {
	       int vi = face->VIndxs()[pv];
	       int ni = face->NIndxs()[pv];

	       Iaka = face->ka*Ia;
	       Ipkd = face->kd*Ip;
	       Ipks = face->ks*Ip;

	       /* P = the current vertex point */
	       P(1, 1) = solid.Vertices()(vi, 1);
	       P(1, 2) = solid.Vertices()(vi, 2);
	       P(1, 3) = solid.Vertices()(vi, 3);

	       /* N = the normal at the current vertex */
	       N(1, 1) = solid.Normals()(ni, 1);
	       N(1, 2) = solid.Normals()(ni, 2);
	       N(1, 3) = solid.Normals()(ni, 3);

	       L = Unitize(lsource - P);
	       V = Unitize(viewpoint - P);
	  
	       NdotL = DotProd3(N, L);
	       R = 2.0*NdotL*N - L;
	       RdotV = DotProd3(R, V);

	       if (NdotL < 0.0) 
		    NdotL = 0.0;
		    
	       if (RdotV < 0.0)
		    RdotV = 0.0;

	       face->Colors()[pv] = Iaka + Ipkd*NdotL + Ipks*pow(RdotV, n);
	  }
     }

     return solid;
}
