#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "Matrix.H"
#include "Solid.H"
#include "List.H"
#include "halftone.h"
#include "render.h"

typedef struct _etentry {
     double dx, dz, x, z, i, di;
     int xmin, ymin, xmax, ymax, face;
} ETentryRec, *ETentry;

int sort_edge(const void *, const void *);

void PrintEdge(ETentry e1)
{
     printf("min = (%d, %d), max = (%d, %d), dx = %6.3f,dz = "
	    "%6.3f, face = %d\n",
	    e1->xmin, e1->ymin, e1->xmax,
	    e1->ymax, e1->dx, e1->dz, e1->face);
}

void PrintEdgeList(List *aet)
{
     printf("Edge list:\n");
     for (List *p = aet; !p->Lastp(); p = p->Next())
	  PrintEdge(ETentry(p->Data()));
}

void RenderSolid(Solid& solid, int xsize, int ysize)
{
     Matrix zbuf(xsize, ysize), I(xsize, ysize);
     SolidFace *face;
     List et[ysize], aet;
     double x1, y1, x2, y2, z1, z2;
     int i, j, l, FullShade;

     /* This is a hack */
     FullShade = (solid.Face(0)->NumNorms() != 1);

     /* Initialize the Z buffer */
     for (i = 1; i <= xsize; i++)
     for (j = 1; j <= ysize; j++) {
	  zbuf(i, j) = 1;
	  I(i, j) = 100.0;
     }
     
     /* Build the edge table */
     for (i=0; i < solid.NumFaces(); i++) {
	  face = solid.Face(i);
	  const int *p = face->VIndxs();
	  l = face->NumVerts();
	  
	  for (j=0; j < l; j++) {
	       x1 = solid.Vertices()(p[j], 1);
	       y1 = solid.Vertices()(p[j], 2);
	       z1 = solid.Vertices()(p[j], 3);
	       x2 = solid.Vertices()(p[(j+1) % l], 1);
	       y2 = solid.Vertices()(p[(j+1) % l], 2);
	       z2 = solid.Vertices()(p[(j+1) % l], 3);

	       if (int(y2) - int(y1) == 0)
		    continue;
	       
	       ETentry entry = new ETentryRec;
	       if (y1 < y2) {
		    entry->x = x1;
		    entry->z = z1;
		    entry->xmin = int(x1);
		    entry->ymin = int(y1);
		    entry->xmax = int(x2);
		    entry->ymax = int(y2);
		    if (FullShade) {
			 entry->i = face->Colors()[j];
			 entry->di =
			      (face->Colors()[(j+1) % l] - entry->i) /
				   (y2 - y1);
#ifdef DEBUG
			 printf("Face %d, edge indx (%d,%d), Imin = %6.2f,"
				" Imax=%6.2f, di=%6.2f\n", i, p[j],
				p[(j+1) % l], entry->i,
				face->Colors()[(j+1) % l], entry->di);
#endif
		    }
	       } else {
		    entry->x = x2;
		    entry->z = z2;
		    entry->xmax = int(x1);
		    entry->ymax = int(y1);
		    entry->xmin = int(x2);
		    entry->ymin = int(y2);
		    if (FullShade) {
			 entry->i = face->Colors()[(j+1) % l];
			 entry->di =
			      (face->Colors()[j] - entry->i) /
				   (y1 - y2);
#ifdef DEBUG
			 printf("Face %d, edge indx (%d,%d), Imin = %6.2f,"
				" Imax=%6.2f, di=%6.2f\n", i, p[(j+1)%l],
				p[j], face->Colors()[(j+1) % l],
				entry->i, entry->di);
#endif
		    }
	       }
	       entry->dx = (x2 - x1) / (y2 - y1);
	       entry->dz = (z2 - z1) / (y2 - y1);
	       entry->face = i;

	       if (y1 < y2)
 		    et[int(y1)].InsertHere((void *) entry);
	       else
		    et[int(y2)].InsertHere((void *) entry);
	  }
     }

     /* Render into the I(mage) matrix and Z buffer */
     
     /* For each scan line y, */
     for (int y = 0; y < ysize; y++) {

	  /* Add new edges to the AET, */
	  for (List *l = &et[y]; l->Data(); l = l->Next()) {
	       const ETentry entry = ETentry(l->Data());
	       aet.InsertSorted(l->Data(), sort_edge);
	  }

	  /* Remove completed edges from the AET. */
	  for (l = &aet; !l->Lastp();) {
	       if (ETentry(l->Data())->ymax == y) {
		    l->Remove();
	       }
	       else
		    l = l->Next();
	  }

	  /* For each pair of edges in each polygon in the AET, */
	  List *p1, *p2;
	  for (p1 = &aet, p2 = p1->Next(); p1->Data();
	       p1 = p2->Next(), p2 = p1->Next()) { 

	       const ETentry e1 = ETentry(p1->Data());
	       const ETentry e2 = ETentry(p2->Data());

	       /* Render the pixels between them. */
	       if (int(e2->x) != int(e1->x)) {
		    double z1 = e1->z, z2 = e2->z;
		    double dzp = (z2 - z1) / (e2->x - e1->x);
		    double i1 = e1->i, i2 = e2->i;
		    double dip = (i2 - i1) / (e2->x - e1->x);

		    for (int x = int(ceil(e1->x)); x < e2->x; x++,
			 z1 += dzp, i1 += dip) {
			 if (z1 < zbuf(x+1, y)) {
			      zbuf(x+1, y) = z1;
			      if (FullShade)
				   I(x+1, y) = i1;
			      else
				   I(x+1, y) =
					solid.Face(e1->face)->Colors()[0];
			 }
		    }
	       }
	       e1->x += e1->dx;
	       e1->z += e1->dz;
	       e2->x += e2->dx;
	       e2->z += e2->dz;
	       if (FullShade) {
		    e1->i += e1->di;
		    e2->i += e2->di;
	       }
	  }
     }

     /* Halftone the image onto the screen */
     FS_Halftone(I, 100.0);
}

int sort_edge(const void *a, const void *b)
{
   int retval;

   // these are ints, so comparing is easy
   retval = (((ETentry)a)->face - ((ETentry)b)->face);
   if (retval) return(retval);

   // we want to cast each to int and subtract here, because
   // the x's should be considered equal if they are the same value
   // when rendered
   retval = int(((ETentry)a)->x) - int(((ETentry)b)->x);
   if (retval) return(retval);

   // here, we want to return -1,1 because the dx has a high resolution
   return((((ETentry)a)->dx > ((ETentry)b)->dx)?1:-1);
}
