Procedures in the source language are transformed into a tuple
containing the body of the procedure and the environment in which it
was created.  The procedure body is transformed to expect an
environment as an argument, instead of a value, and references to the
formal parameter are replaced by a function that extracts the argument
value from the environment explicitly.

Environments are represented as tuples in which the first element is
the value for the formal parameter for the current function and the
second element is the function's environment.

Procedure objects are applied to arguments by extracting the body and
environment from the tuple explicitly and evaluting the body in a new
environment in which the formal parameter (in the source language) is
bound to the argument value.  The formal parameter is "bound" by
placing the argument value in the first element of the environment,
where the function will look for it.

The desugarings are as follows.  [ENV] is intended to be an identifer
that is legal in the source language but not in the object language.
If there is no such identifier, an unused identifer could be found.

\D[[I]] = I
\D[[L]] = L
\D[[(if E_1 E_2 E_3)]] = (if (\D E_1) (\D E_2) (\D E_3))
\D[[(tuple E^*)]] = (tuple (\D E^*))
\D[[(nth E_1 E_2)]] = (nth (\D E_1) (\D E_2))
\D[[(let ((I_1 E_1) ... (I_n E_n)) E_body)]] = standard deguaring for let

\D[[(lambda I E)]] =
(tuple
 (lambda [ENV]
   \D[[[(nth [ENV] 0)/I][(nth [ENV] 1)/[ENV]] E]])
 current-environment)

\D[[(E_1 E_2)]] = ((nth (\D E_1) 0) (tuple (\D E_2) (nth (\D E_1) 1)))

The real work is performed in the lambda desugaring.  Two
substitutions are performed on the procedure body.  The first,
replacing [ENV] with (nth [ENV] 1), allows lexical scoping.  Consider
the example 

   (lambda x
     (lambda y
       x))

The first level transformation of this function (using just the second
substitution) is

   (lambda x
     (\D (lambda y (nth [ENV] 0))))

When the second lambda is tranformed, [ENV] has to refer to the
parent's environment.  However, the argument passed as y will contain
that environment one level further down in the tuple the first
substitution allows it to be accessed.

(define substitute
  (lambda (new id exp)
    (match exp
      ((variable-reference->exp~ id1)
       (if (sym=? id id1)
	   new
	   exp))
      ((application->exp~ rator rand)
       (application->exp (substitute new id rator)
			  (substitute new id rand)))
      ((abstraction->exp~ formal body)
       (if (sym=? id formal)
	   exp
	   (let ((body-free-ids (free-identifiers body)))
	     (if (not (set-member? id body-free-ids))
		 exp
		 (let ((new-free-ids (free-identifiers new)))
		   (if (not (set-member? formal new-free-ids))
		       (abstraction->exp formal
					  (substitute new id body))
		       ; Else alpha renaming necessary
		       (let ((fresh (symbol-not-in-set 
				     (set-union (set-singleton id)
						(set-union body-free-ids 
                                                           new-free-ids)))))
			 (abstraction->exp 
			  fresh
			  (substitute 
                           new 
			   id
			   (substitute 
			    (variable-reference->exp fresh) ; tricky
			    formal
			    body))))))))))
      )))

(define ENV-VAR (fresh-var))

(define transform
  (lambda (exp)
    (let ((xformed (xform (eliminate-let exp) '())))
      xformed)
    ))

(define xform
  (lambda (exp env)
    (write-string "xform: ")
    (write-sexp exp)
    (write-string "\n")
    (match exp
      ((abstraction->exp~ sym body) (xform-abstraction sym body env))
      ((conditional->exp~ pred con alt) (xform-conditional pred con alt env))
      ((tuple->exp~ @args) (xform-tuple args env))
      ((nth->exp~ tuple index) (xform-nth tuple index env))
      ((application->exp~ rator rand) (xform-application rator rand env))
    
;      (`(lambda ,sym ,body) (xform-abstraction sym body env))
;      (`(if ,pred ,con ,alt) (xform-conditional pred con alt env))
;      (`(tuple ,@args) (xform-tuple args env))
;      (`(nth ,tuple ,index) (xform-nth tuple index env))
;      (`(,rator ,rand) (xform-application rator rand env))

      (_ exp))
    ))

(define xform-application
  (lambda (rator rand env)
    (application-exp
     (nth->exp (xform rator env) 0)
     (tuple->exp (xform rand env) (nth->exp (xform rator env) 1)))
    ))

(define xform-nth
  (lambda (tuple index env)
    (nth->exp (xform tuple env) (xform index env))
    ))

(define xform-tuple
  (lambda (args env)
    (tuple->exp (map (lambda (exp) (xform exp env)) args))
    ))

(define xform-abstraction
  (lambda (sym exp env)
    (tuple->exp
     (abstraction->exp
      'ENV-VAR
      (xform
       (substitute (nth->exp 'ENV-VAR 0) sym
		   (substitute (nth->exp 'ENV-VAR 1) 'ENV-VAR
			       exp))
       env))
     (tuple->exp env))
    ))

(define xform-conditional
  (lambda (pred con alt env)
    (conditional->exp (xform pred env) (xform con env) (xform alt env))
    ))

;(define xform-application
;  (lambda (rator rand env)
;    `((nth ,(xform rator env) 0) (tuple ,(xform rand env)
;					(nth ,(xform rator env) 1)))
;    ))
;
;(define xform-nth
;  (lambda (tuple index env)
;    `(nth ,(xform tuple env) ,(xform index env))
;    ))
;
;(define xform-tuple
;  (lambda (args env)
;    `(tuple ,(map (lambda (exp) (xform exp env)) args))
;    ))
;
;(define xform-abstraction
;  (lambda (sym exp env)
;    `(tuple
;      (lambda 'ENV-VAR
;	,(xform
;	  (substitute '(nth 'ENV-VAR 0) ,sym
;		      (substitute '(nth 'ENV-VAR 1) 'ENV-VAR
;				  exp))
;	  env))
;      ,env)
;    ))
;
;(define xform-conditional
;  (lambda (pred con alt env)
;    `(if ,(xform pred env) ,(xform con env) ,(xform alt env))
;    ))
;
